"""
Class of which object is used to plot data
"""
from typing import List, Optional, Union
import matplotlib.text
from matplotlib import pyplot as pl
from matplotlib.transforms import Bbox
from PySide2.QtCore import QTimer, Qt
from PySide2 import QtCore, QtWidgets
from PySide2.QtWidgets import QWidget, QApplication, QTextBrowser

from sohstationviewer.conf import constants
from sohstationviewer.view.util.color import set_color_mode
from sohstationviewer.view.util.functions import (
    get_total_miny_maxy, get_index_from_time
)
from sohstationviewer.view.plotting.plotting_widget.plotting_axes import (
    PlottingAxes
)
from sohstationviewer.view.plotting.plotting_widget.plotting import Plotting
from sohstationviewer.view.save_plot_dialog import SavePlotDialog

from sohstationviewer.controller.plotting_data import format_time
from sohstationviewer.controller.util import display_tracking_info


class PlottingWidget(QtWidgets.QScrollArea):
    """
    Widget to plot different type of channels on a figure and different
        events to serve user's purpose.
    """

    def __init__(self, parent: Union[QWidget, QApplication],
                 tracking_box: QTextBrowser,
                 name: str,
                 main_window: QApplication) -> None:
        """
        :param parent: widget that contains this plotting
            widget
        :param tracking_box: widget to display tracking info
        :param name: name of the plotting widget to keep track of what
            widget the program is working on
        :param main_window: Main window that keep all parameters set by user
        """
        self.parent = parent
        self.main_window = main_window
        self.name = name
        self.tracking_box = tracking_box
        # =============== declare attributes =======================
        """
        processing_log: [(str,str] - list of processing info and type
        """
        self.processing_log = []
        """
        width_base_px: float - width base in pixel which fit the plotting area
        """
        self.width_base_px = 1546.
        """
         width_base: float - basic width of plotting which can be enlarged to
             the maximum of 1 which is 4x of width_base
         """
        self.width_base = 0.25
        """
        plotting_l_base: float - The basic left of plotting area
        """
        self.plotting_l_base = 0.04
        """
        time_ticks_total: int - maximum of total major ticks label displayed
        """
        self.time_ticks_total = 5
        """
        min_x: float - left limit of a channel plot
        max_x: float - right limit of a channel plot
        """
        self.min_x = 0
        self.max_x = 0
        """
        ratio_w: float - the ratio for plotting_w. This value changed when the
            width of the scroll area changed or user choose a different
            scaling ratio
        """
        self.ratio_w = 1.
        """
        plotting_w: float - plotting area's width which is set with ratio_w of
            width_base
        """
        self.plotting_w = self.width_base * self.ratio_w
        """
        plotting_l: float - plotting area's left which is set with ratio_w of
            plotting_l_base.
        TODO: check to see if should keep this left
            unchanged or changed with width of the scrolling area
        """
        self.plotting_l = self.plotting_l_base * self.ratio_w
        """
        plotting_bot: float - bottom of a current plot, decrease by plot_h
            return from self.get_height() whenever a new plot is added
        plotting_bot_pixel: float - bottom of current plot in pixel
        """
        self.plotting_bot = constants.BOTTOM
        self.plotting_bot_pixel = constants.BOTTOM_PX
        """
        display_color: dict - that defined colors to be used.
        """
        self.display_color = {}
        """
        c_mode: str - 'B'/'W' - color modes black or white
        """
        self.c_mode = 'B'
        """
        font_size: float - font size on plot. With some require bigger font,
            +2 to the font_size
        """
        self.base_font_size = 7
        self.font_size = 7
        """
        bottom: float - y position of the bottom edge of all plots in self.axes
            This is to identify the bottom of rulers
        """
        self.bottom = 0
        """
        gap_bar: matplotlib.axes.Axes - axes where gaps are plotted
        """
        self.gap_bar = None
        """
        timestamp_bar_top: matplotlib.axes.Axes - axes to show times that
            places at the top of all channel plots
        timestamp_bar_bottom: matplotlib.axes.Axes - axes to show times that
            places at the bottom of all channel plots
        """
        self.timestamp_bar_top = None
        self.timestamp_bar_bottom = None
        """
        date_mode: str - format of display date
        """
        self.date_mode = 'YYYY-MM-DD'
        """
        peer_plotting_widgets: [PlottingWidget,]: list of all PlottingWidget
            objects that need to be synchronized
        """
        self.peer_plotting_widgets = [self]
        """
        plot_total: int - total of plots display on this widget
        """
        self.plot_total = 0
        """
        axes: [matplotlib.axes.Axes, ] - list of axes for plotting channels
        """
        self.axes = []
        """
        ruler: ConnectionPatch - a line go from top to bottom to help
            user see the corresponding positions of different channels.
            It shows up at the mouse x position when  Ctrl/Cmd + left click,
            and disappears when zoom_marker1 shows up or when ESC is pressed.
        """
        self.ruler = None
        """
        zoom_marker1: ConnectionPatch - a line go from top to bottom to help
            mark where the zooming area start. It shows up at the mouse x
            position when Shift + left click, and disappears when plot has
            zoomed in or when ESC is pressed.
        """
        self.zoom_marker1 = None
        """
        zoom_marker2: ConnectionPatch - a line go from top to bottom to help
            mark where the zooming area end. It shows up at the mouse x
            position when  Shift + left click after zoom_marker1 shown up,
            and disappears when plot has zoomed in or when ESC is pressed.
        """
        self.zoom_marker2 = None
        """
        zoom_marker1_shown: bool - flag for sohview to know if zoom_marker1 is
            shown/hidden
        """
        self.zoom_marker1_shown = False
        """
        follower: int - connection id to help keeping zoom_marker2 following
            mouse move
        """
        self.follower = None
        """
        plotting_data1: dict that includes 'times', 'data', 'ax'-
            first set of data for plotting. It can be either
            DataTypeModel.__init__.soh_data[key] or
            DataTypeModel.__init__.waveform_data[key]['read_data']
        """
        self.plotting_data1 = {}
        """
        plotting_data2: dict that includes 'times' and 'data'-
            second set of data for plotting. It is
            DataTypeModel.__init__.mass_pos_data[key]
        """
        self.plotting_data2 = {}
        # ----------------------------------------------------------------

        QtWidgets.QScrollArea.__init__(self)
        """
        main_widget: QWidget - widget in the scroll area to keep the drawing
            canvas
        """
        self.main_widget = QtWidgets.QWidget(parent)
        """
        plotting_axes: object that helps creating axes for plotting
        """
        self.plotting_axes = PlottingAxes(self, main_window)
        self.plotting_axes.canvas.setParent(self.main_widget)

        self.setWidget(self.main_widget)

        """
        plotting: object that helps with different types of plotting channels
        """
        self.plotting = Plotting(self, self.plotting_axes, main_window)
        """
        new_min_x: store the new minimum time for zooming; used to fix a
        problem where after the first zoom marker is chosen, any ruler that is
        to the left of the first zoom marker does not have text above it.
        """
        self.new_min_x = float('-inf')
        """
        ruler_text: the current text shown above the ruler
        """
        self.ruler_text: Optional[matplotlib.text.Text] = None

        self.set_colors('B')

    # ======================================================================= #
    #                                  EVENT
    # ======================================================================= #
    def resizeEvent(self, event):
        """
        OVERRIDE Qt method.
        When plottingWidget's viewport is resized along with parent window
            (even opening MainWindow triggers resizeEvent too), self.ratio_w,
            self.plotting_w, self.plotting_l will be recalculated so that the
            drawn channels fit inside the width of the new viewport.
        If self.plot_total == 0, when there is no channels drawn, height
            is set to the height of viewport to cover the whole viewport since
            viewport and widget may be in different colors.

        :param event: QResizeEvent - resize event
        """
        geo = self.maximumViewportSize()

        # set view size fit with the scroll's view port size
        self.main_widget.setFixedWidth(geo.width())
        self.ratio_w = geo.width() / self.width_base_px
        self.font_size = self.ratio_w * self.base_font_size
        self.plotting_w = self.ratio_w * self.width_base
        self.plotting_l = self.ratio_w * self.plotting_l_base
        if self.plot_total == 0:
            self.main_widget.setFixedHeight(geo.height())

        return super(PlottingWidget, self).resizeEvent(event)

    def get_timestamp(self, event):
        """
        Get timestamp which is xdata value from the coordinate data of the
            event. Normally, xdata is an attribute of event. However, when
            mouse falls into spaces in between axes, event.xdata is None.

        :param event: button_press_event - mouse click event
        """
        x, y = event.x, event.y  # coordinate data
        inv = self.axes[0].transData.inverted()
        # convert to timestamp, bc event.xdata is None in the space bw axes
        xdata = inv.transform((x, y))[0]
        return xdata

    def zoom_bw_markers(self, xdata):
        """
        Zoom the section in between zoom_marker1 and zoom_marker2:
            + set flag zoom_marker1_shown to False
            + set lim to the new curr_min_x and new curr_max_x set by x of
                zoom_marker1 and zoom_marker2
            + hide zoom_marker1 and zoom_marker2

        :param xdata: float - time value in plot
        """
        if self.new_min_x == xdata:
            return
        self.zoom_marker1_shown = False
        [self.min_x, self.max_x] = sorted([self.new_min_x, xdata])
        self.set_lim()
        self.zoom_marker1.set_visible(False)
        self.zoom_marker2.set_visible(False)
        self.plotting_axes.canvas.draw()

    def on_pick_event(self, event):
        """
        When click mouse on a clickable data point (dot with picker=True in
            Plotting),
        + Point's info will be displayed in tracking_box
        + If the chan_data has key 'logIdx', raise the Search Messages dialog,
            focus SOH tab, roll to the corresponding line.
        """
        artist = event.artist
        ax = artist.axes
        chan_id = ax.chan
        if isinstance(artist, pl.Line2D):
            chan_data = self.plotting_data1[chan_id]
            # list of x values of the plot
            x_list = artist.get_xdata()
            # list of y values of the plot
            y_list = artist.get_ydata()

            # index of the clicked point on the plot
            click_plot_index = event.ind[0]

            # time, val of the clicked point
            clicked_time = x_list[click_plot_index]
            clicked_val = y_list[click_plot_index]

            list_idx, section_idx = get_index_from_time(
                chan_data, clicked_time, clicked_val)

            if list_idx is None:
                display_tracking_info(self.tracking_box, "Point not found.")
                return

            clicked_data = chan_data['data'][list_idx][section_idx]
            if hasattr(ax, 'unit_bw'):
                clicked_data = ax.unit_bw.format(clicked_data)
            formatted_clicked_time = format_time(
                clicked_time, self.date_mode, 'HH:MM:SS')
            info_str = (f"<pre>Channel: {chan_id}   "
                        f"Point:{click_plot_index + 1}   "
                        f"Time: {formatted_clicked_time}   "
                        f"Value: {clicked_data}</pre>")
            display_tracking_info(self.tracking_box, info_str)
            if 'logIdx' in chan_data.keys():
                # For Reftek, need to hightlight the corresponding
                # SOH message line based on the log_idx of the clicked point
                self.parent.search_message_dialog.show()
                clicked_log_idx = chan_data['logIdx'][0][section_idx]
                try:
                    self.parent.search_message_dialog. \
                        show_log_entry_from_data_index(clicked_log_idx)
                except ValueError as e:
                    QtWidgets.QMessageBox.warning(self, "Not found",
                                                  str(e))

    def on_button_press_event(self, event):
        """
        When click mouse on the current plottingWidget, SOHView will loop
            through different plottingWidgets to do the same task for
            interaction:
            + shift+click: call on_shift_click() to do zooming. This is
                disregarded in TimePowerSquareWidget because it isn't subjected
                to be zoomed in.
            + ctrl+click or cmd+click in mac: call on_ctrl_cmd_click() to show
                ruler

        :param event: button_press_event - mouse click event
        """
        for w in self.peer_plotting_widgets:
            if w.is_working:
                # try to zoom or use ruler while not done plotting
                return

        modifiers = event.guiEvent.modifiers()
        if self.plotting_data1 == {}:
            return
        if self.name == 'TPS':
            if modifiers == QtCore.Qt.ShiftModifier:
                # not start zooming from TPS Widget
                return
            # on_button_press_event() take place after on_pick_event where
            # tps_t was assigned in TPS Widget
            xdata = self.tps_t
        else:
            xdata = self.get_timestamp(event)

        # We only want to remove the text on the ruler when we start zooming in
        # or move the ruler to another location.
        if modifiers in [QtCore.Qt.ControlModifier, QtCore.Qt.MetaModifier,
                         Qt.ShiftModifier]:
            try:
                self.ruler_text.remove()
                self.ruler_text = None
            except AttributeError:
                pass

        for w in self.peer_plotting_widgets:
            if modifiers in [QtCore.Qt.ControlModifier, QtCore.Qt.MetaModifier,
                             Qt.ShiftModifier]:
                try:
                    w.ruler_text.remove()
                    w.ruler_text = None
                except AttributeError:
                    pass
            if modifiers == QtCore.Qt.ShiftModifier:
                w.on_shift_click(xdata)
            elif modifiers in [QtCore.Qt.ControlModifier,
                               QtCore.Qt.MetaModifier]:
                w.on_ctrl_cmd_click(xdata)
            w.draw()

        self.parent.show()
        self.parent.activateWindow()
        self.parent.raise_()

    def on_ctrl_cmd_click(self, xdata):
        """
        On ctrl/cmd + left click
            + hide zoom_marker1
            + connect ruler to follow mouse
            + show ruler

        :param xdata: float - time value of a channel plot
        """
        self.zoom_marker1.set_visible(False)
        self.zoom_marker1_shown = False
        try:
            self.fig.canvas.mpl_disconnect(self.follower)
        except AttributeError:
            pass
        self.set_ruler_visibled(self.ruler, xdata)
        if xdata >= self.min_x:
            ruler_text_content = format_time(xdata, self.parent.date_format,
                                             'HH:MM:SS')
            self.ruler_text = self.plotting_axes.fig.text(
                xdata, 75, ruler_text_content,
                verticalalignment='top',
                horizontalalignment='center',
                transform=self.timestamp_bar_top.transData,
                color=self.display_color['time_ruler'],
                size=self.timestamp_bar_top.get_xticklabels()[0].get_fontsize()
            )

    def on_shift_click(self, xdata):
        """
        On shift + left click:
        if zoom_marker1 not shown yet:
            + hide ruler
            + connect zoom_marker2 to follow mouse
            + show zoom_marker1
        else:
            + show zoom_marker2
            + zoom data in between 2 zoomMarkers

        :param xdata: float - time value of a channel plot
        """
        if not self.zoom_marker1_shown:
            self.ruler.set_visible(False)
            self.set_ruler_visibled(self.zoom_marker1, xdata)
            self.new_min_x = xdata
            self.zoom_marker1_shown = True
        else:
            self.set_ruler_visibled(self.zoom_marker2, xdata)
            self.plotting_axes.canvas.draw()
            QTimer.singleShot(1, lambda: self.zoom_bw_markers(xdata))
            # self.zoom_bw_markers(xdata)

    def set_ruler_visibled(self, ruler, xdata):
        """
        When set ruler visibled, x of positions of top and bottom have to
            follow the given x.
        If ruler is zoom_marker2, set it to follow the x of mouse's position.

        :param ruler: ConnectionPatch - one of ruler, zoom_marker1,
            zoom_marker2 defined in __init__
        :param xdata: float - time value of a channel plot
        """

        ruler.set_visible(True)
        ruler.xy1 = (xdata, 0)
        ruler.xy2 = (xdata, self.bottom)

    def keyPressEvent(self, event):
        """
        OVERRIDE Qt method.
        When press on Escape key, hide all rulers and set False for
            zoom_marker1_shown on all plotting widgets

        :param event: QKeyEvent - event to know what key is pressed
        """
        if event.key() == QtCore.Qt.Key_Escape:
            try:
                self.ruler_text.remove()
                self.ruler_text = None
            except AttributeError:
                pass
            for w in self.peer_plotting_widgets:
                w.set_rulers_invisible()
                w.zoom_marker1_shown = False
                try:
                    w.ruler_text.remove()
                    w.ruler_text = None
                except AttributeError:
                    pass
                w.draw()
        return super(PlottingWidget, self).keyPressEvent(event)

    def set_rulers_invisible(self):
        """
        Hide ruler, zoom_marker1, zoom_marker2
        """
        self.ruler.set_visible(False)
        self.zoom_marker1.set_visible(False)
        self.zoom_marker2.set_visible(False)

    def check_gap_related_to_range(self, gap: List[float],
                                   min_x: float, max_x: float):
        """
        Check if the given gap inside either completely or partly of range
            (min_x, max_x)
        :param gap: a list of a gap's start time and end time
        :param min_x: minimum of range
        :param max_x: maximum of range
        :return: True if the gap is related to the range, False otherwise.
        """
        return (self.min_x <= min(gap) <= self.max_x
                or self.min_x <= max(gap) <= self.max_x
                or min(gap) <= self.min_x <= max(gap)
                or min(gap) <= self.max_x <= max(gap))

    def set_lim(self, first_time=False):
        """
        + Re-decimate channel data to get more detail when zoom in
            (get_zoom_data)
        + Update timestamp bar with new ticks info
        + Update gap_bar by setting xlim and re-calculate gap total.
        + for each axes, set new x_lim, y_lim, label for totals of data points

        :param first_time: bool - flag shows that this set_lim is called the
            fist time for this data set or not.
        """
        self.plotting_axes.update_timestamp_bar(self.timestamp_bar_top)
        self.plotting_axes.update_timestamp_bar(self.timestamp_bar_bottom)
        if self.gap_bar is not None:
            self.gap_bar.set_xlim(self.min_x, self.max_x)
            if not first_time:
                new_gaps = [g for g in self.plotting_axes.gaps
                            if self.check_gap_related_to_range(g,
                                                               self.min_x,
                                                               self.max_x)]

                # reset total of samples on the right
                self.gap_bar.sampleLbl.set_text(len(new_gaps))

        for ax in self.axes:
            if hasattr(ax, 'x') and ax.x is None:
                # the plot_none bar is at the end, no need to process
                break

            ax.set_xlim(self.min_x, self.max_x)
            if not first_time:
                new_min_y = None
                new_max_y = None
                if hasattr(ax, 'x_list'):
                    if not hasattr(ax, 'y_list'):
                        # dotForTime plots have attribute 'x_list' but not
                        # 'y_list'
                        continue
                    total_points = 0
                    tr_min_ys = []
                    tr_max_ys = []
                    for x, y in zip(ax.x_list, ax.y_list):
                        if len(x) == 0:
                            continue
                        if self.min_x > x[-1] or self.max_x < x[0]:
                            continue
                        ret = get_total_miny_maxy(x, y, self.min_x, self.max_x)
                        total_tr_points, tr_min_y, tr_max_y = ret
                        total_points += total_tr_points
                        if tr_min_y is not None:
                            tr_min_ys.append(tr_min_y)
                            tr_max_ys.append(tr_max_y)
                    if tr_min_ys != []:
                        new_min_y = min(tr_min_ys)
                        new_max_y = max(tr_max_ys)
                else:
                    total_points = len(ax.x)
                    if hasattr(ax, 'y') and len(ax.y) > 0:
                        new_min_y = min(ax.y)
                        new_max_y = max(ax.y)
                try:
                    ax.sampleLbl.set_text(total_points)
                except AttributeError:
                    # for case of having top and bottom total points
                    # which is for RT130's SOH only, trust in total point
                    # calculated in set_axes_info
                    pass

                if new_min_y is not None:
                    self.plotting_axes.set_axes_ylim(
                        ax, new_min_y, new_max_y, ax.chan_db_info)

    def draw(self):
        """
        Update drawing on the widget.
        """
        try:
            self.plotting_axes.canvas.draw()
            # a bug on mac:
            # not showing updated info until clicking on another window
            # fix by calling repaint()
            self.main_widget.repaint()
        except TypeError:
            pass

    # ######## Functions for outside world #####
    def init_size(self):
        """
        Set widget to fit with the view port when window first open.
        """
        geo = self.maximumViewportSize()
        if self.plot_total == 0:
            # set view size fit with the scroll's view port size
            self.main_widget.setFixedWidth(geo.width())
            self.main_widget.setFixedHeight(geo.height())

    def set_colors(self, mode):
        """
        Set color mode, colors' definition, background color for the graphic.

        param mode: str - 'B'/'W' - color modes black or white
        """
        self.c_mode = mode
        self.display_color = set_color_mode(mode)
        self.plotting_axes.fig.patch.set_facecolor(
            self.display_color['background'])

    def set_peer_plotting_widgets(self, widgets):
        """
        Set other plotting widgets so that their rulers and zoom task can work
            synchronously.
        :param widgets: [PlottingWidgets,] - List of other plotting widgets
        """
        self.peer_plotting_widgets = widgets

    def save_plot(self, default_name='plot'):
        if self.c_mode != self.main_window.color_mode:
            main_color = constants.ALL_COLOR_MODES[self.main_window.color_mode]
            curr_color = constants.ALL_COLOR_MODES[self.c_mode]
            msg = (f"Main window's color mode is {main_color}"
                   f" but the mode haven't been applied to plotting.\n\n"
                   f"Do you want to cancel to apply {main_color} mode "
                   f"by clicking RePlot?\n"
                   f"Or continue with {curr_color}?")
            msgbox = QtWidgets.QMessageBox()
            msgbox.setWindowTitle("Color Mode Conflict")
            msgbox.setText(msg)
            msgbox.addButton(QtWidgets.QMessageBox.Cancel)
            msgbox.addButton('Continue', QtWidgets.QMessageBox.YesRole)
            result = msgbox.exec_()
            if result == QtWidgets.QMessageBox.Cancel:
                return
            self.main_window.color_mode = self.c_mode
            if self.c_mode == 'B':
                self.main_window.background_black_radio_button.setChecked(True)
            else:
                self.main_window.background_white_radio_button.setChecked(True)
        if self.c_mode == 'B':
            msg = ("The current background mode is black.\n"
                   "Do you want to cancel to change the background mode "
                   "before saving the plots to file?")
            msgbox = QtWidgets.QMessageBox()
            msgbox.setWindowTitle("Background Mode Confirmation")
            msgbox.setText(msg)
            msgbox.addButton(QtWidgets.QMessageBox.Cancel)
            msgbox.addButton('Continue', QtWidgets.QMessageBox.YesRole)
            result = msgbox.exec_()
            if result == QtWidgets.QMessageBox.Cancel:
                return
        save_plot_dlg = SavePlotDialog(
            self.parent, self.main_window, default_name)
        save_plot_dlg.exec_()
        save_file_path = save_plot_dlg.save_file_path
        if save_file_path is None:
            return
        dpi = save_plot_dlg.dpi

        self.plotting_axes.fig.savefig(
            save_file_path,
            bbox_inches=Bbox([[0, self.plotting_bot*100],
                              [self.ratio_w*15.5, 100]]),
            dpi=dpi
        )
        msg = f"Graph is saved at {save_file_path}"
        display_tracking_info(self.tracking_box, msg)

    def clear(self):
        self.plotting_axes.fig.clear()
        self.axes = []
        self.plot_total = 0
        self.init_size()
        self.draw()
